import sys
import base64
import logging
import copy

import rsb
from rsb.converter import PredicateConverterList, SchemaAndByteArrayConverter
from rsb import Event

from rsb.protocol.introspection.Hello_pb2 import Hello
from rsb.protocol.introspection.Bye_pb2 import Bye


class Forwarder(SchemaAndByteArrayConverter):

    def serialize(self, data):
        if isinstance(data, tuple):
            return data[1], data[0]
        else:
            return data, self.wireSchema


class Bridge(object):
    basic_types = {'integer': int, 'float': float, 'string': str, 'bool': bool}

    def __init__(self, rsb_scope, rsb_conf, wamp, message_type):
        logging.info("register scopes:")
        self.rsb_scope = rsb_scope
        self.wamp_scope = rsb_scope[1:].replace('/', '.')
        self.converter = None
        self.skipNext = False;
        logging.info("RSB Scope %s" % self.rsb_scope)
        logging.info("WAMP Scope is %s" % self.wamp_scope)
        self.wamp = wamp
        self.wamp_listener = self.wamp.subscribe(self.on_wamp_message, self.wamp_scope)
        if message_type in Bridge.basic_types:
            self.wamp_callback = self.send_primitive_data
            self.rsb_callback = self.on_primitive_message
            self.rsb_type = Bridge.basic_types[message_type]
        else:
            self.wamp_callback = self.send_rst
            self.rsb_callback = self.on_bytearray_message
        self.rsb_listener = rsb.createListener(self.rsb_scope)
        self.rsb_listener.addHandler(self.rsb_callback)
        self.rsb_publisher = rsb.createInformer(self.rsb_scope)

    def on_bytearray_message(self, event):
        if 'wamp' in event.metaData.userInfos:
            logging.debug("received OWN rsb bytearray on %s, skipping..." % self.rsb_scope)
            return
        logging.debug('received rsb bytearray on %s' % self.rsb_scope)
        logging.debug('event length %d' % len(event.data))
        logging.debug('sent to %s' %self.wamp_scope)
        try:
            msg = '\0' + base64.b64encode(event.data).decode('ascii')
            self.wamp.publish(self.wamp_scope, msg)
        except Exception as e:
            print e
            sys.exit(1)

    def on_primitive_message(self, event):
        if 'wamp' in event.metaData.userInfos:
            logging.debug("received OWN rsb primitive on %s, skipping..." % self.rsb_scope)
            return
        logging.debug("received rsb primtive [%s] on %s" % (str(event.data), self.rsb_scope))
        logging.debug("sent to %s" % self.wamp_scope)
        self.wamp.publish(self.wamp_scope, self.rsb_type(event.data))

    def send_rst(self, event):
        try:
            logging.info("send rst message to %s" % self.rsb_scope)
            binary_data = bytearray(base64.b64decode(event[1:]))
            ev = Event(scope=self.rsb_publisher.getScope(),
                       data=binary_data,
                       type=tuple)
            self.rsb_publisher.publishData(binary_data, userInfos={'wamp':''})
        except Exception as e:
            print e
            sys.exit(1)

    def send_primitive_data(self, event):
        try:
            logging.info("send primitive message [%s] message to %s" % (str(event),self.rsb_scope))
            self.rsb_publisher.publishData(self.rsb_type(event),
                                           userInfos={'wamp':''})
        except Exception as e:
            print e
            sys.exit(1)

    def on_wamp_message(self, event):
        logging.debug('received wamp message on %s' % self.wamp_scope)
        self.wamp_callback(event)

    def release(self):
        self.rsb_listener.deactivate()
        if self.rsb_publisher is not False:
            self.rsb_publisher.deactivate()


class SessionHandler(object):

    def __init__(self, wamp_session, log_level=logging.WARNING):
        logging.basicConfig(level=log_level)

        self.wamp_session = wamp_session
        rsb_conf = copy.deepcopy(rsb.getDefaultParticipantConfig())
        trans = rsb_conf.getTransports()
        conv = Forwarder()
        conv_list = PredicateConverterList(bytearray)
        conv_list.addConverter(conv,
                               wireSchemaPredicate=lambda wireSchema: not wireSchema.startswith('rsb.protocol.introspection'),
                               dataTypePredicate=lambda dataType: dataType == tuple)

        # register introspection types
        for clazz in [Hello, Bye]:
            converter = rsb.converter.ProtocolBufferConverter(messageClass=clazz)
            conv_list.addConverter(converter)

        for t in trans:
            t.converters = conv_list

        rsb.setDefaultParticipantConfig(rsb_conf)

    def register_scope(self, rsb_scope, message_type):
        logging.info("trying to register on scope %s with message type %s" %
                     (rsb_scope, message_type))

        # supress rsb logging warnings
        logging.getLogger("rsb").setLevel(logging.ERROR)
        if rsb_scope not in self.scopes.keys():
            b = Bridge(rsb_scope, self.rsb_conf, self.wamp_session, message_type)
            self.scopes[rsb_scope] = b
            return "Scope registered"
        return "Scope already exists"



import logging
import os

try:
    import asyncio
except ImportError:
    # Trollius >= 0.3 was renamed
    import trollius as asyncio

from os import environ
from autobahn.asyncio.wamp import ApplicationSession, ApplicationRunner
from services import SessionHandler


class Component(ApplicationSession):

    @staticmethod
    def on_ping(event):
        logging.debug(event)

    @asyncio.coroutine
    def onJoin(self, details):
        # init members
        if os.environ.get('DEBUG') in ['1','True','true','TRUE']:
            log_level = logging.DEBUG
        else:
            log_level = logging.WARN
        logging.basicConfig(level=log_level)
        session = SessionHandler(self, log_level)

        # register RPC
        reg = yield self.register(session.register_scope, 'service.displayserver.register')

        # setup ping
        sub = yield self.subscribe(self.on_ping, "com.wamp.ping")

        print 'kogniserver(asyncio) started...'

        try:
            while True:
                logging.debug("ping")
                self.publish("com.wamp.ping", "ping")
                yield asyncio.sleep(1)
        except Exception:
            print("shutting kogniserver...")
            for scope in self.scopes.values():
                scope.release()

def main_entry():
    from autobahn.asyncio.wamp import ApplicationRunner
    runner = ApplicationRunner(url = u"ws://127.0.0.1:8181/ws", realm = u"realm1")
    runner.run(Component)

if __name__ == '__main__':
    main_entry()


from twisted.internet.defer import inlineCallbacks

from autobahn.twisted.util import sleep
from autobahn.twisted.wamp import ApplicationSession
from autobahn.wamp.exception import ApplicationError

import sys
import logging

from services import SessionHandler

class AppSession(ApplicationSession):

    @staticmethod
    def on_ping(event):
        print event

    @inlineCallbacks
    def onJoin(self, details):

        # init members
        logging.basicConfig()
        session = SessionHandler(self)

        # register RPCs
        reg = yield self.register(session.register_scope, 'service.displayserver.register')
        sub = yield self.subscribe(self.on_ping, "com.wamp.ping")
        try:
            while True:
                try:
                    print "ping"
                    self.publish("com.wamp.ping", "ping")
                except ApplicationError as e:
                    print e
                    sys.exit(1)
                yield sleep(1)

        except Exception:
            print "shut down server..."
            for scope in self.scopes.values():
                scope.release()

def main_entry():
    from autobahn.twisted.wamp import ApplicationRunner
    runner = ApplicationRunner(url = u"ws://127.0.0.1:8181/ws", realm = u"realm1")
    runner.run(AppSession)

if __name__ == '__main__':
    main_entry()


from setuptools import setup, find_packages
from setuptools.command.install import install
import os
from os.path import join, exists, abspath
import json


class Install(install):
    user_options = install.user_options + [
            ('protopath=', None, 'path to protocol files'),
            ('overwrite-config', None, 'overwrite previously installed crossbar config')
        ]

    def initialize_options(self):
        install.initialize_options(self)
        self.protopath = False
        self.overwrite_config = False

    def run(self):
        install.run(self)
        install.do_egg_install(self)
        self.prefix = abspath(self.prefix)
        self.protopath = abspath(self.protopath) if self.protopath else False
        conf_path = join(self.prefix, 'etc/crossbar')
        if exists(conf_path) is False:
             os.makedirs(conf_path)

        conf_file = join(conf_path, 'config.json')
        if exists(conf_file) is False or self.overwrite_config:
            protopath = self.protopath if self.protopath else join(self.prefix, 'share/rst0.12/proto')
            with open(conf_file, 'w') as target:
                try:
                    with open('config.json.in') as template:
                        j = json.load(template)
                        paths = j['workers'][0]['transports'][0]['paths']
                        paths['/']['directory'] = join(self.prefix, paths['/']['directory'])
                        paths['proto']['directory'] = protopath
                        json.dump(j, target, indent=4)
                except Exception as e:
                    raise Exception(abspath('config.json.in'))

setup(
    name='kogniserver',
    version='0.1.1',
    maintainer='Alexander Neumann',
    maintainer_email='aleneum@gmail.com',
    url='http://github.com/aleneum/kogniserver',
    description="Interface server of the KogniHome project",
    platforms=['Any'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    tests_require=['nose>=1.3', 'coverage'],
    install_requires=['txaio', 'pyasn1', 'autobahn<0.13.0', 'crossbar<0.13.0', 'trollius', 'rsb-python<0.13.0'],
    entry_points={
        "console_scripts": [
            "kogniserver = kogniserver.async:main_entry",
        ]
    },
    cmdclass={
        'install': Install
    },
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Communications',
        'Topic :: Home Automation',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking'
    ],
)
